#
# 引数レゾルバ
#

"""
テンプレート引数 (:mod:`blueprintpy.core.argument` に格納する値を解決・代入するパッケージ.

インタフェース :class:`.Resolver` を基底クラスとし、カスタムテンプレート内で拡張することができます.
"""

from . import numbers, string, lists
from .base import Resolver as Resolver



#
# デフォルト引数ハンドラ
#

"""
引数ハンドラのデフォルト実装
"""

from typing import List, Optional, Type

from blueprintpy.core import Argument
from blueprintpy.cli.args_handler import ArgsHandlerBase

from .exceptions import ValidationError
from .resolver import Resolver


class DefaultArgsHandler(ArgsHandlerBase):

    """ (プロパティ :code:`__handler_name__` の値は :code:`__default__` です.)
    """

    __handler_name__ = "__default__"

    @staticmethod
    def handle_args(args: List[Argument]) -> List[Argument]:
        """
        各テンプレート引数について、プロパティ :code:`argtype` の値に対応する名称のレゾルバを探します.
        次に各レゾルバに引数オブジェクトを投入し、処理結果を プロパティ :code:`value` に代入して返します.

        対応するレゾルバが存在しない場合は :code:`RuntimeError`, レゾルバの処理が失敗した場合は :class:`.exceptions.ValidationError` が送出されます.
        """
        prepared_args: List[Argument] = []

        # 各引数について
        for arg in args:
            # 対応するresolverを探し
            argtype = arg.argtype if arg.argtype is not None else "str"
            resolver = DefaultArgsHandler.__search_resolver(argtype)
            if resolver is None:
                raise RuntimeError(f"No resolver registered that can handle argument type \033[36m{argtype}\033[0m!")

            while True:
                # 適当に名前と説明文を表示したのち
                print(f"\033[32;1m{arg.name}\033[0m ({arg.description}):")

                try:
                    # resolveして追加
                    prepared_args.append(resolver.resolve(arg))
                    break
                except ValidationError as e:
                    print(f"\033[31mError\033[0m {e.argument.name}: {e.reason}")
                    continue

        return prepared_args

    @staticmethod
    def __search_resolver(argtype: str) -> Optional[Type[Resolver]]:
        resolver_candidates = list(filter(lambda resolver: resolver.__resolver_type__ == argtype, Resolver.resolvers))
        return resolver_candidates[0] if len(resolver_candidates) == 1 else None



#
# 展開先に書き込む準備ができたContent
#
from pathlib import Path
from typing import Optional


class PreparedContent:
    """展開可能なContent
    """

    def __init__(self,
                 dest_path: Path,
                 source: Optional[Path] = None,
                 extract_object: Optional[bytes] = None) -> None:
        """
        Args:
            dest_path (Path): 展開先の絶対パス
            source (Optional[Path], optional): 展開するファイルの絶対パス
            extract_object (Optional[bytes], optional): 展開するオブジェクトのバイト列

        Raises:
            ValueError: 展開するべきコンテンツが指定されなかった場合

        Note:
            :code:`source`, :code:`extract_object` のいずれかに有効な値が入力されている必要があります.
        """

        self.dest_path = dest_path

        if source is None and extract_object is None:
            raise ValueError("Invalid argument")

        self.source = source
        self.extract_object = extract_object



#
# 引数ハンドラ基底クラス
#

"""
テンプレート引数ハンドラの基底クラス
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import List, Set, Type

from blueprintpy.core import Argument


class ArgsHandlerBase(metaclass=ABCMeta):
    """
    Attributes:
        handlers (Set[Type[ArgsHandlerBase]]) : 基底クラスを継承したArgsHandlerのセット
        __handler_name__ (str): 引数ハンドラを特定するための名前
    """

    handlers: Set[Type[ArgsHandlerBase]] = set()
    __handler_name__: str = ""

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        cls.handlers.add(cls)

    @staticmethod
    @abstractmethod
    def handle_args(args: List[Argument]) -> List[Argument]:
        """与えられたテンプレート引数リストに適切な値を挿入します。

        Args:
            args (List[Argument]): 対象のテンプレート引数リスト

        Returns:
            List[Argument]: 挿入結果
        """



#
# Contentを適切な場所に配置する
#

from shutil import copyfile

from .prepared_content import PreparedContent


class ContentExtractor:
    """PreparedContentの展開
    """

    @staticmethod
    def extract(content: PreparedContent):
        """引数に渡されたPreparedContentを適切な場所に配置します.

        Args:
            content (PreparedContent): 配置するコンテンツ

        Raises:
            ValueError: コンテンツのパスが不正であるか、コンテンツの指すオブジェクトが不正だった場合

        Note:
            コンテンツの各パスは絶対パスである必要があります.

            また、:code:`content` のパスは検証されません.
            内容のバリデーションは引数ハンドリングプロセスで行うことを推奨します.
        """

        # 読み込み元、配置先が絶対パスであることを確認する
        if not content.dest_path.is_absolute():
            raise ValueError("Invalid argument")
        if content.source is not None and not content.source.is_absolute():
            raise ValueError("Invalid argument")

        # 書き込み先パスに到達できるよう、親ディレクトリまでは作成しておく
        content.dest_path.parent.mkdir(parents=True, exist_ok=True)

        # 書き込み
        if content.source is not None:
            copyfile(content.source, content.dest_path)
        elif content.extract_object is not None:
            with open(content.dest_path, "wb") as f:
                f.write(content.extract_object)
        else:
            raise ValueError("Invalid argument")



#
# コンフィグローダ
#
import importlib
from json.decoder import JSONDecoder
from typing import Optional

from blueprintpy.core import Argument, Config, Content


class ConfigLoader:
    """テンプレート構成ローダ
    """

    @staticmethod
    def load(config_json_str: str) -> Config:
        """json形式のテンプレート構成情報からテンプレート構成を生成します.

        Args:
            config_json_str (str): json形式のテンプレート構成情報

        Raises:
            KeyError: 構成情報に必要なプロパティが含まれていなかった場合.
            ValueError: 構成情報に不正な値が含まれていた場合.
            ModuleNotFoundError, ImportError: 引数ハンドラのインポートに失敗した場合.

        Returns:
            Config: 生成結果
        """

        # jsonデコード
        config_json = JSONDecoder().decode(config_json_str)

        # ルートに name, args, contentsがあることを期待する
        try:
            name, args, contents = tuple([config_json[key] for key in ['name', 'args', 'contents']])
        except KeyError:
            raise KeyError("name, args, and contents are required as root property of json.")

        # args, contentsはそれぞれインスタンスに起こす
        try:
            arg_instances = [Argument(**arg) for arg in args]
            content_instances = [Content(**content) for content in contents]
        except TypeError:
            raise ValueError("unexpected parameter of initilaize Argument or Content")

        # args_handlerが存在するならモジュールをロードする この時点でArgsHandlerBase.handlerからアクセスできるようになっているはず
        args_handler_path: Optional[str] = config_json['args_handler'] if 'args_handler' in config_json else None
        if args_handler_path is not None:
            importlib.import_module(args_handler_path)

        # Config生成
        return Config(name, arg_instances, content_instances, args_handler_path)



#
# blueprintpy CLI
#

"""
blueprintpy CLIツール
"""

from .cli import main


def blueprint() -> int:
    """
    コマンド :code:`blueprint` を提供する関数. KeyboardInterrupt (:code:`^C`) のみがこの関数でハンドルされ,
    それ以外の処理は 関数 :func:`.cli.main` が担当しています.

    Returns:
        int: 終了コード
    """

    try:
        return main()
    except KeyboardInterrupt:
        print("\n")
        print("\033[31;1mProcess aborted by user interaction.\033[0m")
        print("\n")
        return 1



#
# コンフィグシリアライザ
#
from functools import reduce
from typing import Any, Dict

from blueprintpy.core import Config


class ConfigSerializer:
    """テンプレート構成シリアライザ
    """

    @staticmethod
    def serialize(config: Config) -> Dict[str, Any]:
        """テンプレート構成オブジェクトをシリアライズし、辞書形式に変換します.

        Args:
            config (Config): シリアライズする構成オブジェクト

        Returns:
            Dict[str, Any]: シリアライズ結果
        """

        serialized: Dict[str, Any] = {}

        # 名前はそのまま
        serialized['name'] = config.name

        # Argumentは適当に処理する valueはコンストラクタにないのでスルー
        serialized_args = [
            reduce(
                lambda p, n: p | n,
                list(
                    [{key: getattr(arg, key)}
                     if getattr(arg, key) is not None else{} for key in ['name', 'description', 'argtype', 'default_value']]))
            for arg in config.args]
        serialized['args'] = serialized_args

        # ハンドラはそのまま ただし存在しなければ追加しない
        if config.args_handler_name is not None:
            serialized['args_handler'] = config.args_handler_name

        # Contentも同様
        serialized_contents = [{'src': content.src, 'dest': content.dest} for content in config.contents]
        serialized['contents'] = serialized_contents

        return serialized



#
# 整数・浮動小数点数
#
"""
数値形式のデータを代入するresolver
"""

from blueprintpy.core import Argument

from ..exceptions import ValidationError

from .base import Resolver


class IntResolver(Resolver):
    """__resolver_type__: :code:`int`
    """

    __resolver_type__ = "int"

    @staticmethod
    def resolve(argument: Argument) -> Argument:
        """
        整数形式のデータについて、コンソールからの入力を元に値を生成します.
        """
        value = input("> ")

        if value != "":
            try:
                argument.value = int(value)
                return argument
            except ValueError:
                raise ValidationError(argument, f"Invalid value that cannot be processed as an integer")

        if argument.default_value is None:
            raise ValidationError(argument, "valid value required")

        argument.value = argument.default_value
        return argument


class FloatResolver(Resolver):
    """__resolver_type__: :code:`float`
    """

    __argtype__ = "float"

    @staticmethod
    def resolve(argument: Argument) -> Argument:
        """
        浮動小数点形式のデータについて、コンソールからの入力を元に値を生成します.
        """

        value = input("> ")

        if value != "":
            try:
                argument.value = float(value)
                return argument
            except ValueError:
                raise ValidationError(argument, f"Invalid value that cannot be processed as an integer")

        if argument.default_value is None:
            raise ValidationError(argument, "valid value required")

        argument.value = argument.default_value
        return argument



#
# 単純な文字列
#
"""
文字列形式のデータを代入するresolver
"""

from blueprintpy.core import Argument

from ..exceptions import ValidationError
from .base import Resolver


class StringResolver(Resolver):
    """__resolver_type__: :code:`str`
    """

    __resolver_type__ = "str"

    @staticmethod
    def resolve(argument: Argument) -> Argument:
        """
        文字列形式のデータについて、コンソールからの入力を元に値を生成します.
        """

        value = input("> ")
        if value != "":
            argument.value = value
            return argument

        if argument.default_value is None:
            raise ValidationError(argument, "valid value required")

        argument.value = argument.default_value
        return argument



#
# テンプレート引数
#

from typing import Any, Optional


class Argument:
    """パッケージテンプレート内で展開される引数.
    """

    def __init__(self,
                 name: str,
                 description: str,
                 argtype: Optional[str] = None,
                 default_value: Optional[str] = None) -> None:
        """
        Args:
            name (str): 引数名
            description (str): 引数の説明
            argtype (Optional[str], optional): 引数タイプ
            default_value (Optional[str], optional): デフォルト値

        Note:
            :code:`argtype` パラメータは引数の型を示します.
            blueprintpy コアモジュールでは利用されませんが、CLIツールのデフォルト引数ハンドラでは対応する名称の
            レゾルバ (:mod:`blueprintpy.cli.args_handler.resolver`) を呼び出そうとします.
        """

        self.name = name
        self.description = description
        self.argtype = argtype
        self.default_value = default_value
        self.value: Any = None



#
# テンプレート構成
#

from typing import List, Optional

from .argument import Argument
from .content import Content


class Config:
    """パッケージテンプレートの構成.
    """

    def __init__(self,
                 name: str,
                 args: List[Argument],
                 contents: List[Content],
                 args_handler_name: Optional[str] = None) -> None:
        """
        Args:
            name (str): テンプレート名
            args (List[Argument]): テンプレート引数リスト
            contents (List[Content]): 展開されるファイルコンテンツのリスト
            args_handler_name (Optional[str], optional): 引数ハンドラの名前

        Note:
            :code:`args_handler_name` パラメータには、引数ハンドラ(:mod:`blueprintpy.cli.args_handler`)の
            属性 :code:`__handler_name__` の値を指定してください.
            指定がなかった場合は、blueprintpy.cliが自動でデフォルトのハンドラを呼び出します.
        """
        self.name = name
        self.args = args
        self.contents = contents
        self.args_handler_name = args_handler_name



#
# blueprintpy CLI
#
"""
blueprintpy CLIツールのメインモジュール
"""

import importlib
import sys
from argparse import ArgumentParser
from pathlib import Path
from types import ModuleType
from typing import Optional

from blueprintpy.core import ContentBuilder, ContentExtractor

from .args_handler import ArgsHandlerBase
from .config_loader import ConfigLoader


def main() -> int:
    """コマンド :code:`blueprint` の実装.コマンドライン引数を処理し、テンプレート展開処理を実行します.

    Returns:
        int: 終了コード
    """
    # コマンドライン引数の設定
    parser = ArgumentParser(
        prog='blueprint',
        usage="%(prog)s [target] [--name name_of_template] [--template_dir path/to/template]",
        description="Python package template extractor")
    parser.add_argument(
        "target",
        nargs="?",
        default=".",
        help="template extract destination (Defaults to current directory)")
    parser.add_argument(
        "--name", "-n",
        default="default",
        help="name of template (Defaults to library internal template \"default\")")
    parser.add_argument(
        "--template_dir", "-t",
        help="template root directory (optional)")
    parser.add_argument(
        "--dry_run", "-d",
        action='store_true',
        help="shows only which files are expanded")

    # パースして情報を取得
    args = parser.parse_args()
    extract_root: Path = Path(args.target).absolute()
    template_name: str = args.name
    additional_template_dir: Optional[Path] = Path(args.template_dir).absolute() if args.template_dir is not None else None
    is_dry_run = args.dry_run

    if is_dry_run:
        print("Execute as dry-run mode.")

    # template_rootが指定された場合はsys.pathに追加しておく
    if additional_template_dir is not None:
        sys.path.append(str(additional_template_dir))
    # そうでなければ ~/.blueprintpy/をsys.pathに追加する
    # このディレクトリは存在してもしなくてもよい
    else:
        sys.path.append(str(Path.home() / ".blueprintpy"))

    # パスを解決し、テンプレートをimportする
    # blueprintpy.internal_templates、blueprintpy_templatesでそれぞれ該当する名前のテンプレートを探す
    template_module: Optional[ModuleType] = None
    try:
        template_module = importlib.import_module(f"blueprintpy.internal_templates.{template_name}")
    except ImportError:
        print(f"template \033[36m{template_name}\033[0m not found at inrernal template. found custom template directory...")

    # どちらにもなければエラー
    if template_module is None:
        try:
            template_module = importlib.import_module(f"blueprintpy_templates.{template_name}")
        except ImportError:
            print("\033[31;1mfailed to import template! check if the path is valid.\033[0m")
            print(f"search path: \033[36mblueprintpy_templates.{template_name}\033[0m")
            return 1

    # importしたモジュールからテンプレートの親ディレクトリを特定し、template.jsonを読み込む
    template_root = Path(template_module.__file__).parent
    with open(template_root / "template.json") as f:
        template_json = f.read()
    config = ConfigLoader.load(template_json)

    # 引数ハンドラを特定する
    args_handler_name = config.args_handler_name or "__default__"
    args_handler_candidates = list(filter(lambda handler: handler.__handler_name__ == args_handler_name, ArgsHandlerBase.handlers))
    if len(args_handler_candidates) != 1:
        print("\033[31;1mcould not identify argument handler!\033[0m")
        return 1
    args_handler = args_handler_candidates[0]

    # 引数ハンドラに渡して値をセットしてもらう
    prepared_args = args_handler.handle_args(config.args)

    # Contentをビルド
    content_builder = ContentBuilder(template_root, extract_root, prepared_args)
    prepared_contents = [content_builder.build(content) for content in config.contents]

    # 配置
    for content in prepared_contents:
        if not is_dry_run:
            ContentExtractor.extract(content)
        else:
            print(f"expand: {content.source} -> {content.dest_path}")

    # 完了
    print("Succeeded.")
    return 0


if __name__ == "__main__":
    result = 0
    try:
        result = main() or 0
    except KeyboardInterrupt:
        print("Ctrl+C")
        exit(result)



#
# 単純なリスト
#
"""
リスト形式のデータを代入するresolver
"""

from typing import List

from blueprintpy.core import Argument

from .base import Resolver


class ArrayResolver(Resolver):
    """__resolver_type__: :code:`array`
    """

    __resolver_type__ = "array"

    @staticmethod
    def resolve(argument: Argument) -> Argument:
        """
        リスト形式のデータについて、コンソールからの入力を元に値を生成します.
        入力するたびに内部リストに値が蓄積され、 :code:`EOF` を送信することで値が確定されます.

        Note:
            現在のバージョンでは **デフォルト値を考慮した動作をしない** ことに注意してください.
            何らかの値が必ず要求されます.
        """

        value_buffer: List[str] = []
        print("Entered 'list mode'. ^D (Ctrl+D) to exit.")
        while True:
            try:
                value = input("> ")
            except EOFError:
                print()
                break

            if value != "":
                value_buffer.append(value)

            print(f"Current: {','.join(value_buffer)}")

        argument.value = value_buffer
        return argument



#
# Python package template initialzer
#

"""
blueprintpy コアパッケージ
"""

from .argument import Argument as Argument
from .config import Config as Config
from .content import Content as Content
from .content_builder import ContentBuilder as ContentBuilder
from .content_extractor import ContentExtractor as ContentExtractor
from .prepared_content import PreparedContent as PreparedContent



#
# 展開されるファイルコンテンツ
#

import os
from typing import Optional


class Content:
    """パッケージテンプレートに含まれるファイル.
    """

    def __init__(self,
                 src: str,
                 dest: str,
                 use_template_engine: Optional[bool] = None) -> None:
        """
        Args:
            src (str): オリジナル相対パス
            dest (str): コピー先相対パス
            use_template_engine (Optional[bool], optional): テンプレートエンジンを使用するか

        Note:
            :code:`use_template_engine` にNoneが渡された場合, 拡張子が :code:`.j2` または :code:`.jinja2` のファイルではTrue、
            それ以外のファイルではFalseがセットされます.
        """

        self.src = src
        self.dest = dest

        if use_template_engine is not None:
            self.use_template_engine = use_template_engine
        else:
            _, suffix = os.path.splitext(src)
            self.use_template_engine = suffix in ['.j2', '.jinja2']



#
# 引数レゾルバの基底クラス
#

"""
引数レゾルバの基底クラス
"""

from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Set, Type

from blueprintpy.core import Argument


class Resolver(metaclass=ABCMeta):
    """
    Attributes:
        resolvers (Set[Type[Resolver]]) : 基底クラスを継承したResolverのセット
        __resolver_type__ (str): 引数レゾルバを特定するための名前
    """

    resolvers: Set[Type[Resolver]] = set()
    __resolver_type__: str = ""

    def __init_subclass__(cls) -> None:
        super.__init_subclass__()

        cls.resolvers.add(cls)

    @staticmethod
    @abstractmethod
    def resolve(argument: Argument) -> Argument:
        """与えられたテンプレート引数に適切な値を代入します.

        Args:
            argument (Argument): 対象のテンプレート引数

        Returns:
            Argument: 代入結果
        """
        raise NotImplementedError()



#
# 引数ハンドリング例外
#
"""引数の処理過程で発生する例外
"""

from blueprintpy.core import Argument


class ArgumentHandlingError(Exception):
    """(基底クラス)
    """


class ValidationError(ArgumentHandlingError):
    """引数に渡された値が不正だった場合に発生する例外.
    """

    def __init__(self, argument: Argument, reason: str) -> None:
        self.argument = argument
        self.reason = reason
        super().__init__(argument, reason)



#
# Content引数展開
#
from functools import reduce
from pathlib import Path
from typing import Any, Dict, List, Optional

from jinja2 import BaseLoader, Environment, FileSystemLoader

from .argument import Argument
from .content import Content
from .prepared_content import PreparedContent


class ContentBuilder:
    """Contentビルダ
    """

    def __init__(self,
                 template_root: Path,
                 extract_root: Path,
                 template_args: List[Argument],
                 template_loader: Optional[BaseLoader] = None) -> None:
        """
        Args:
            template_root (Path): 読み込むテンプレートのルート
            extract_root (Path): 展開先のルート
            template_args (List[Argument]): テンプレート引数
            template_loader (Optional[BaseLoader], optional): テンプレートローダー

        Note:
            :code:`template_root` , :code:`extract_root` は、インスタンス初期化時に絶対パスに変換されます.
        """

        loader = template_loader if template_loader is not None else FileSystemLoader(template_root)
        self.env = Environment(loader=loader, trim_blocks=True)
        self.template_root = template_root.absolute()
        self.extract_root = extract_root.absolute()

        # テンプレート引数はdictに変換
        template_parameter_dicts: List[Dict[str, Any]] = [{arg.name: arg.value} for arg in template_args]
        if len(template_parameter_dicts) == 0:
            # (functools.reduceは空っぽのシーケンスに使うと怒られる)
            self.template_args: Dict[str, Any] = {}
            return

        self.template_args = reduce(lambda prev, next: prev | next, template_parameter_dicts)

    def build(self, content: Content) -> PreparedContent:
        """引数に与えられたContentオブジェクトを読み込んで処理し、展開可能なContentオブジェクトを生成します.

        Args:
            content (Content): 変換元のContent

        Returns:
            PreparedContent: 処理結果
        """

        # Content.destをjinjaに通し、配置先の絶対パスを生成
        dest_template = Environment(loader=BaseLoader()).from_string(content.dest)
        formatted_dest = dest_template.render(self.template_args)
        dest_abspath = self.extract_root / formatted_dest

        # テンプレートエンジンを通さなくてよければそのまま戻る
        if not content.use_template_engine:
            source_abspath = self.template_root / content.src
            return PreparedContent(dest_abspath, source_abspath, None)

        # srcが指すファイルを読み込み、jinjaに通して返す
        content_template = self.env.get_template(content.src)
        render_result = content_template.render(self.template_args).encode()
        return PreparedContent(dest_abspath, None, render_result)
